"""
Parser for LS-Dyna
"""

import re

HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD})*"
NONWS = r"\S+"
INT = r"\d+"
PROGRAM = r"[A-Z0-9\-_]+"
EXPIRATION_DATE = r"((0[1-9]|1[012])[\/](0[1-9]|[12][0-9]|3[01])[\/](19|20)\d\d)"
CPUS_USED = r"[\d|-]"

PROGRAM_LINE = (
    rf"(?P<program>{PROGRAM})\s+"
    rf"(?P<expiration_date>{EXPIRATION_DATE})\s+"
    rf"(?P<used>{CPUS_USED})\s+ "
    rf"(?P<free>{INT})\s+ "
    rf"(?P<max>{INT})\s+\|\s+"
    rf"(?P<queue>{INT})"
)

USAGE_LINE = (
    rf"\s+" rf"(?P<user>\S+)\s+" rf"(?P<port>{INT})\@" rf"(?P<lead_host>{HOSTNAME})\s+" rf"(?P<used>{INT})"
)

GROUP_LINE = (
    rf"\s+LICENSE GROUP\s+"
    rf"(?P<used>{INT})\s+"
    rf"(?P<free>{INT})\s+"
    rf"(?P<max>{INT})\s\|\s+"
    rf"(?P<queue>{INT})"
)

RX_PROGRAM = re.compile(PROGRAM_LINE)
RX_USAGE = re.compile(USAGE_LINE)
RX_TOTAL = re.compile(GROUP_LINE)


def parse_program_line(line: str):
    """
    Parse the program line in the LS-Dyna output.
    Data we need:
    - ``program``: license name
    - ``max``: total amount of licenses in the server
    The ``used``value will be evaluated later.
    """
    parsed_program = RX_PROGRAM.match(line)
    if parsed_program is None:
        return None
    program_data = parsed_program.groupdict()
    program_data["program"] = program_data["program"].lower()

    return {
        "program": program_data["program"],
        "total": int(program_data["max"]),
    }


def parse_usage_line(line: str):
    """
    Parse the usage line in the LS-Dyna output.
    Data we need:
    - ``user_name``: user who booked the license
    - ``lead_host``: host using the license
    - ``booked``: quantity of licenses being used

    Obs: this line doesn't include the license name.
    The license in use is the last one parsed before this line.
    """
    parsed_usage = RX_USAGE.match(line)
    if parsed_usage is None:
        return None
    usage_data = parsed_usage.groupdict()

    return {
        "user_name": usage_data["user"],
        "lead_host": usage_data["lead_host"],
        "booked": int(usage_data["used"]),
    }


def parse_total_line(line: str):
    """
    Parse the total line in the LS-Dyna output.
    Data we need:
    - ``used``: quantity of licenses being used in the server

    Obs: all LS-Dyna licenses in the server share the same
    license pool, where the amount of available licenses can be consumed
    by any license in the server.

    The ``used`` value is the sum of all licenses in use in the pool.
    Since we calculate the available quantity as ``available = total - used``,
    we'll use the group ``used`` value for all licenses, so the available
    will reflect the correct amount of free licenses in the pool.
    """
    parsed_total = RX_TOTAL.match(line)
    if parsed_total is None:
        return None
    total_data = parsed_total.groupdict()

    return int(total_data["used"])


def parse(s: str):
    """
    Parse the LS-Dyna output, using regex to match the lines we need:
    - ``program line``: info about each license
    - ``usage line``: info about users using licenses
    - ``total line``: info about the pool, which affects all licenses

    Since the usage line doesn't have the name of the license in use,
    we're saving each parsed license in a list. This way, we can find
    which license the user is using by checking the last parsed license
    in this list.

    Also, all the licenses in the server will have their ``used`` value
    filled with the group used value, since it'll reflect the correct
    amount of free licenses in the pool.
    """
    parsed_data: dict = {}
    program_list: list = []
    group_used: int = 0

    for line in s.splitlines():
        parsed_program = parse_program_line(line)
        parsed_usage = parse_usage_line(line)
        parsed_total = parse_total_line(line)

        if parsed_program:
            parsed_data[parsed_program["program"]] = {
                "total": int(parsed_program["total"]),
                "uses": [],
            }
            program_list.append(parsed_program["program"])

        elif parsed_usage:
            program = program_list[-1]
            parsed_data[program]["uses"].append(parsed_usage)

        elif parsed_total:
            group_used = parsed_total
            break

    for program in parsed_data.keys():
        parsed_data[program]["used"] = group_used

    return parsed_data



"""
Parser for OLicense
"""
import re
from typing import Optional

HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD})*"
INT = r"\d+"
EXPIRATION_DATE = (
    r"[0-9]{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1]) (2[0-3]|[01][0-9]):[0-5][0-9]:[0-5][0-9]"
)

FEATURE_LINE = (
    r"^\s+(?P<feature>\w+);"
    r"\s+(?P<license_type>\w+);"
    rf"\s+(?P<total>{INT});"
    rf"\s+(?P<expiration_date>{EXPIRATION_DATE});"
)

IN_USE_LINE = rf"^\s+(?P<in_use>{INT})\s+FloatsLockedBy:$"

USAGE_LINE = rf"^\s+(?P<user>\S+)@(?P<lead_host>{HOSTNAME})\s+#(?P<booked>{INT})$"

RX_FEATURE = re.compile(FEATURE_LINE)
RX_IN_USE = re.compile(IN_USE_LINE)
RX_USAGE = re.compile(USAGE_LINE)


def parse_feature_line(line: str) -> Optional[dict]:
    """
    Parse the "feature" line in the OLicense output.
    Data we need:
    - ``feature``: license name
    - ``total``: total amount of licenses

    Obs: OLicense supports more than one license with the same feature name.
    In case a feature has more than one license associated with it,
    the total amount of licenses is the sum of all licenses with the same name.
    """
    parsed_feature = RX_FEATURE.match(line)
    if parsed_feature is None:
        return None
    feature_data = parsed_feature.groupdict()

    return {
        "feature": feature_data["feature"].lower(),
        "total": int(feature_data["total"]),
    }


def parse_in_use_line(line: str) -> Optional[int]:
    """
    Parse the "in use" line in the Olicense output.
    Data we need:
    - ``in_use``: quantity of licenses being used by each user

    Obs: this line doesn't include the license name.
    The license in use is the last one parsed before this line.
    It also doesn't include the user name.
    The user using the license is the next "usage" line parsed after this line.
    The total amount of licenses being used is the sum of all "in use" lines.
    """

    parsed_in_use = RX_IN_USE.match(line)
    if parsed_in_use is None:
        return None
    in_use_data = parsed_in_use.groupdict()

    return int(in_use_data["in_use"])


def parse_usage_line(line: str) -> Optional[dict]:
    """
    Parse the usage line in the Olicense output.
    Data we need:
    -``user_name``: user who booked the license
    -``lead_host``: host using the license
    -``booked``: quantity of licenses booked by the user

    Obs: this line also doesn't include the license name.
    The license in use is the last one parsed before this line.
    """
    parsed_usage = RX_USAGE.match(line)
    if parsed_usage is None:
        return None
    usage_data = parsed_usage.groupdict()

    return {
        "user_name": usage_data["user"],
        "lead_host": usage_data["lead_host"],
        "booked": int(usage_data["booked"]),
    }


def parse(server_output: str) -> dict:
    """
    Parse the OLicense output using regex to match the lines we need:
    -``feature line``: info about each license
    -``in use line``: info about licenses in use
    -``usage line``: info about users using licenses

    Since the "in use" and "usage" lines don't have the name of the license in use,
    we're saving each parsed license in a list. This way, we can find which
    license is being used by checking the last parsed license in the list.

    If a feature has more than one license associated with it, the total amount
    of licenses is the sum of all licenses with the same name.

    Example of output:
        ...
        ftire_adams;         	FreeFloating;	3;	2022-12-31 23:59:59;
        1 FloatsLockedBy:
        sbhyma@RD0087712 #1
        ...
        ftire_adams;         	FreeFloating;	1;	2023-02-28 23:59:00;

    This would be parsed as:
        "ftire_adams": {"total": 4, "used" 1, "uses": {
            "user_name": "sbhyma", "lead_host": "RD0087712", "booked": 1
        }}
    """
    parsed_data: dict = {}
    feature_list: list = []

    for line in server_output.splitlines():
        parsed_feature = parse_feature_line(line)
        if parsed_feature:
            feature = parsed_feature["feature"]
            total = parsed_feature["total"]
            if feature not in feature_list:
                feature_list.append(feature)
                parsed_data[feature] = {"used": 0, "total": total, "uses": []}
            else:
                parsed_data[feature]["total"] += total
            continue

        parsed_in_use = parse_in_use_line(line)
        if parsed_in_use:
            feature = feature_list[-1]
            parsed_data[feature]["used"] += parsed_in_use
            continue

        parsed_usage = parse_usage_line(line)
        if parsed_usage:
            feature = feature_list[-1]
            parsed_data[feature]["uses"].append(parsed_usage)

    return parsed_data



"""
Common utilities for slurm commands.
"""
import os

from lm_agent.logs import logger

SCONTROL_PATH = "/usr/bin/scontrol"
SACCTMGR_PATH = "/usr/bin/sacctmgr"
SQUEUE_PATH = "/usr/bin/squeue"
CMD_TIMEOUT = 5
ENCODING = "UTF8"


def get_job_context():
    """Get and return variables from the job environment."""
    ctxt = dict()
    try:
        ctxt = {
            "cluster_name": os.environ["SLURM_CLUSTER_NAME"],
            "job_id": os.environ["SLURM_JOB_ID"],
            "lead_host": os.environ["SLURM_JOB_NODELIST"].split(",")[0],
            "user_name": os.environ["SLURM_JOB_USER"],
            "job_licenses": os.environ.get("SLURM_JOB_LICENSES", ""),
        }
    except KeyError as e:
        # If not all keys could be assigned, then return non 0 exit status
        logger.error(
            f"All required environment variables were not set, missing: {e}. "
            "Expecting: SLURM_CLUSTER_NAME, SLURM_JOB_ID, SLURM_JOB_NODELIST, "
            "SLURM_JOB_USER, SLURM_JOB_LICENSES"
        )

    return ctxt



"""Utilities that interact with slurm."""
import asyncio
import re
import shlex
import subprocess
from typing import Dict, List, Optional, Union

from lm_agent.backend_utils.models import LicenseBooking
from lm_agent.logs import logger
from lm_agent.workload_managers.slurm.common import (
    CMD_TIMEOUT,
    ENCODING,
    SACCTMGR_PATH,
    SCONTROL_PATH,
    SQUEUE_PATH,
)


class SqueueParserUnexpectedInputError(ValueError):
    """Unexpected squeue output."""


class ScontrolRetrievalFailure(Exception):
    """
    Could not get SLURM data for job id.
    The following function's return code was zero:

    await asyncio.create_subprocess_shell(
        shlex.join([SCONTROL_PATH, "show", f"job={slurm_job_id}"]),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    """


def _match_requested_license(requested_license: str) -> Union[dict, None]:
    license_regex = re.compile(
        r"(?P<product>[a-zA-Z0-9-_]+)\.(?P<feature>[a-zA-Z0-9-_]+)@(?P<server_type>\w+)(:(?P<tokens>\d+))?"
    )

    matches = license_regex.match(requested_license)

    if not matches:
        return None

    groups = matches.groupdict()
    if not groups["tokens"]:
        groups["tokens"] = "1"

    return {
        "product_feature": groups["product"] + "." + groups["feature"],
        "server_type": groups["server_type"],
        "tokens": int(groups["tokens"]),
    }


def get_required_licenses_for_job(job_licenses: str) -> List:
    """Retrieve the required licenses for a job."""

    license_array = job_licenses.split(",")
    logger.debug(f"##### License array for job: {license_array} #####")

    required_licenses: List = []

    if not license_array:
        return required_licenses

    if license_array[0] == "(null)":
        return required_licenses

    for requested_license in license_array:
        matched_license_items = _match_requested_license(requested_license)
        if not matched_license_items:
            continue
        product_feature = matched_license_items["product_feature"]
        booked = matched_license_items["tokens"]

        # Create the license booking
        license_booking = LicenseBooking(
            product_feature=product_feature,
            quantity=booked,
        )
        required_licenses.append(license_booking)

    return required_licenses


async def get_all_features_used_values() -> Optional[Dict[str, int]]:
    """
    Parse the output from `scontrol show lic` and return a dictionary of
    product_feature: used_tokens.

    Note: the line with the counters doesn't include the feature name,
    so to find the feature it's necessary to look at the previous line.
    """
    product_feature_line = re.compile(
        r"LicenseName=(?P<product>[a-zA-Z0-9-_]+).(?P<feature>[a-zA-Z0-9-_]+)@(?P<license_server_type>\w+)"
    )

    used_tokens_line = re.compile(
        r"^\s*Total=(?P<total>\d+) Used=(?P<used>\d+) Free=(?P<free>\d+) Reserved=(?P<reserved>\d+) Remote=(?P<remote>\w+)"  # noqa
    )

    scontrol_output = await scontrol_show_lic()

    parsed_data: dict = {}
    product_feature_list: list = []

    for line in scontrol_output.split("\n"):
        parsed_product_feature = product_feature_line.match(line)
        if parsed_product_feature:
            product_feature_data = parsed_product_feature.groupdict()
            product_feature = product_feature_data["product"] + "." + product_feature_data["feature"]
            product_feature_list.append(product_feature)
            continue

        parsed_used_tokens = used_tokens_line.match(line)
        if parsed_used_tokens:
            used_data = parsed_used_tokens.groupdict()
            product_feature = product_feature_list[-1]
            parsed_data[product_feature] = int(used_data["used"])
            continue

    return parsed_data


async def scontrol_show_lic():
    """
    Get the license usage from scontrol.
    """

    cmd = [
        SCONTROL_PATH,
        "show",
        "lic",
    ]

    proc = await asyncio.create_subprocess_shell(
        shlex.join(cmd), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT
    )

    stdout, _ = await asyncio.wait_for(proc.communicate(), CMD_TIMEOUT)
    output = str(stdout, encoding=ENCODING)
    logger.debug("##### scontrol show lic #####")
    return output


async def get_cluster_name() -> str:
    cmd = [
        SACCTMGR_PATH,
        "list",
        "cluster",
        "-nP",
        "format=Cluster",
    ]
    logger.debug("##### sacctmgr get cluster name cmd #####")

    sacctmgr_modify_resource = await asyncio.create_subprocess_shell(
        shlex.join(cmd),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    cluster_name_stdout, _ = await asyncio.wait_for(
        sacctmgr_modify_resource.communicate(),
        CMD_TIMEOUT,
    )
    return cluster_name_stdout.decode("utf8").strip()


async def sacctmgr_modify_resource(product: str, feature: str, num_tokens) -> bool:
    """
    Update the license resource in slurm.
    """
    cmd = [
        SACCTMGR_PATH,
        "modify",
        "resource",
        f"name={product}.{feature}",
        "set",
        f"count={num_tokens}",
        "-i",
    ]
    logger.debug("##### sacctmgr update cmd #####")
    logger.debug(f"{' '.join(cmd)}")

    sacctmgr_modify_resource = await asyncio.create_subprocess_shell(
        shlex.join(cmd),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    modify_resource_stdout, _ = await asyncio.wait_for(
        sacctmgr_modify_resource.communicate(),
        CMD_TIMEOUT,
    )

    rc = sacctmgr_modify_resource.returncode

    if not rc == 0:
        logger.warning(f"rc = {rc}!")
        logger.warning(modify_resource_stdout)
        return False
    return True


async def get_all_product_features_from_cluster() -> List[str]:
    """
    Returns a list of all product.feature in the cluster.
    """
    show_lic_output = await scontrol_show_lic()

    PRODUCT_FEATURE = r"LicenseName=(?P<product>[a-zA-Z0-9-_]+).(?P<feature>[a-zA-Z0-9-_]+)"
    RX_PRODUCT_FEATURE = re.compile(PRODUCT_FEATURE)

    parsed_features = []
    output = show_lic_output.split("\n")
    for line in output:
        parsed_line = RX_PRODUCT_FEATURE.match(line)
        if parsed_line:
            parsed_data = parsed_line.groupdict()
            product = parsed_data["product"]
            feature = parsed_data["feature"]
            parsed_features.append(f"{product}.{feature}")

    return parsed_features


def return_formatted_squeue_out() -> str:
    """
    Call squeue via Popen and return the formatted output.

    Return the squeue output in the form "<job_id>|<run_time>|<state>".
    """

    result = subprocess.run(
        f"{shlex.quote(SQUEUE_PATH)} --noheader --format='%A|%M|%T'",
        capture_output=True,
        shell=True,
        encoding="utf-8",
    )
    if result.returncode != 0:
        logger.error(result.stderr)
        raise Exception(result.stderr)

    return result.stdout.strip()


def _total_time_in_seconds(time_string: str) -> int:
    """
    Return the runtime in seconds for a job.

    This function takes a slurm time string ("<days>-<hours>:<minutes>:<seconds>") as input, parses
    and converts each of the units in the time string to seconds and returns a computed value, the sum of
    the days, hours, minutes and seconds (in seconds).
    """
    MINUTE = 60
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR

    days = 0
    hours = 0
    splitted_time = [int(value) for value in re.split("-|:", time_string)]
    splitted_time_len = len(splitted_time)

    if splitted_time_len == 4:
        days, hours, minutes, seconds = splitted_time
    elif splitted_time_len == 3:
        hours, minutes, seconds = splitted_time
    else:
        minutes, seconds = splitted_time

    return days * DAY + hours * HOUR + minutes * MINUTE + seconds


def squeue_parser(squeue_formatted_output) -> List:
    """Parse the squeue formatted output."""

    squeue_parsed_output: List = []

    if not squeue_formatted_output:
        return squeue_parsed_output

    def parse_squeue_line():
        """Parse a line from squeue formatted output."""
        try:
            job_id, run_time, state = line.split("|")
        except ValueError as e:
            logger.error(e)
            raise SqueueParserUnexpectedInputError()
        return job_id, run_time, state

    for line in squeue_formatted_output.split():
        job_id, run_time, state = parse_squeue_line()
        squeue_parsed_output.append(
            {
                "job_id": int(job_id),
                "run_time_in_seconds": _total_time_in_seconds(run_time),
                "state": state,
            }
        )

    return squeue_parsed_output



import logging
import sys
from enum import Enum
from pathlib import Path
from typing import Optional

from pydantic import AnyHttpUrl, BaseSettings, Field
from pydantic.error_wrappers import ValidationError

logger = logging.getLogger("lm_agent.config")


DEFAULT_DOTENV_PATH = Path("/etc/default/license-manager-agent")
DEFAULT_CACHE_DIR = Path.home() / Path(".cache/license-manager")
PRODUCT_FEATURE_RX = r"^.+?\..+$"
ENCODING = "UTF8"
TOOL_TIMEOUT = 6  # seconds


class LogLevelEnum(str, Enum):
    """
    Log level name enforcement
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


_DEFAULT_BIN_PATH = Path(__file__).parent.parent / "bin"


class Settings(BaseSettings):
    """
    App config.

    If you are setting these in the environment, you must prefix "LM2_AGENT_", e.g.
    LM2_AGENT_LOG_LEVEL=DEBUG
    """

    DEPLOY_ENV: str = "LOCAL"

    # base url of an endpoint serving the licensemanager2 backend
    BACKEND_BASE_URL: AnyHttpUrl = Field("http://127.0.0.1:8000")

    # location of the log directory
    LOG_BASE_DIR: Optional[str]

    # path to the license server features config file
    LICENSE_SERVER_FEATURES_CONFIG_PATH: Optional[str]

    # path to the binary for lmutil (needed for FlexLM licenses)
    LMUTIL_PATH: Path = _DEFAULT_BIN_PATH

    # path to the binary for rlmutil (needed for RLM licenses)
    RLMUTIL_PATH: Path = _DEFAULT_BIN_PATH

    # path to the binary for lstc_qrun (needed for LS-Dyna licenses)
    LSDYNA_PATH: Path = _DEFAULT_BIN_PATH

    # path to the binary for lmxendutil (needed for LM-X licenses)
    LMXENDUTIL_PATH: Path = _DEFAULT_BIN_PATH

    # path to the binary for olixtool (needed for OLicense licenses)
    OLIXTOOL_PATH: Path = _DEFAULT_BIN_PATH

    # reservation name for reconciliation
    RESERVATION_IDENTIFIER: str = "license-manager-reservation"

    # License Manager user name to create reservations
    LM_USER: str = "license-manager"

    # debug mode turns on certain dangerous operations
    DEBUG: bool = False

    # sentry specific settings
    SENTRY_DSN: Optional[str] = None
    SENTRY_SAMPLE_RATE: Optional[float] = Field(1.0, gt=0.0, le=1.0)

    # log level
    LOG_LEVEL: LogLevelEnum = LogLevelEnum.INFO

    # OIDC config for machine-to-machine security
    OIDC_DOMAIN: str
    OIDC_AUDIENCE: str
    OIDC_CLIENT_ID: str
    OIDC_CLIENT_SECRET: str

    # Token cache directory
    CACHE_DIR: Path = DEFAULT_CACHE_DIR

    # If set to `True`, reconcile will be triggered by Prolog/Epilog. Set to `False` to disable this.
    USE_RECONCILE_IN_PROLOG_EPILOG: bool = True

    class Config:
        env_prefix = "LM2_AGENT_"
        if DEFAULT_DOTENV_PATH.is_file():
            env_file = DEFAULT_DOTENV_PATH
        else:
            env_file = Path(".env")


def init_settings() -> Settings:
    """
    Build SETTINGS, and offer a way to gracefully fail if required settings cannot be loaded

    TODO: Determine if this is true. If Settings is missing a required field, it should throw
          an exception that causes the uvicorn process to exit gracefully (with an error code)
    """
    try:
        return Settings()
    except ValidationError as e:
        logger.error(f"Failed to load settings: {str(e)}")
        # neither fastapi nor uvicorn appear to offer a way to do a graceful
        # shutdown as of now, so this is what we've got.
        sys.exit(1)


settings = init_settings()



#!/usr/bin/env python3
"""
Reconciliation functionality live here.
"""
import asyncio
from typing import Dict, List

from lm_agent.backend_utils.models import BookingSchema
from lm_agent.backend_utils.utils import (
    get_all_features_bookings_sum,
    get_bookings_for_job_id,
    get_cluster_configs_from_backend,
    get_cluster_grace_times,
    get_cluster_jobs_from_backend,
    make_feature_update,
    remove_job_by_slurm_job_id,
)
from lm_agent.exceptions import LicenseManagerEmptyReportError, LicenseManagerReservationFailure
from lm_agent.license_report import report
from lm_agent.logs import logger
from lm_agent.server_interfaces.license_server_interface import LicenseReportItem
from lm_agent.workload_managers.slurm.cmd_utils import (
    get_all_features_used_values,
    return_formatted_squeue_out,
    squeue_parser,
)
from lm_agent.workload_managers.slurm.reservations import (
    scontrol_create_reservation,
    scontrol_delete_reservation,
    scontrol_show_reservation,
    scontrol_update_reservation,
)


def get_greatest_grace_time_for_job(grace_times: Dict[int, int], job_bookings: List[BookingSchema]) -> int:
    """
    Find the greatest grace_time among the features booked by the given job_id.
    """
    greatest_booking_feature = max(
        job_bookings, default=None, key=lambda job_booking: grace_times[job_booking.feature_id]
    )
    if not greatest_booking_feature:
        return -1
    return grace_times[greatest_booking_feature.feature_id]


def get_running_jobs(squeue_result: List) -> List:
    return [j for j in squeue_result if j["state"] == "RUNNING"]


async def clean_jobs_by_grace_time():
    """
    Clean the jobs where running time is greater than the grace_time.
    """
    logger.debug("GRACE_TIME START")

    formatted_squeue_output = return_formatted_squeue_out()
    if not formatted_squeue_output:
        logger.debug("GRACE_TIME no squeue")
        await clean_jobs(None)
        logger.debug("GRACE_TIME cleaned bookings that are not in the queue")
        return

    squeue_result = squeue_parser(formatted_squeue_output)
    squeue_running_jobs = get_running_jobs(squeue_result)

    get_bookings_call = [get_bookings_for_job_id(job["job_id"]) for job in squeue_running_jobs]
    results = await asyncio.gather(*get_bookings_call)
    bookings_for_running_jobs = {job["job_id"]: result for job, result in zip(squeue_running_jobs, results)}

    grace_times = await get_cluster_grace_times()

    # get the grace_time for each job
    for job in squeue_running_jobs:
        slurm_job_id = job["job_id"]
        greatest_grace_time = get_greatest_grace_time_for_job(
            grace_times, bookings_for_running_jobs[slurm_job_id]
        )
        # if the running_time is greater than the greatest grace_time, delete the booking for it
        if job["run_time_in_seconds"] > greatest_grace_time and greatest_grace_time != -1:
            logger.debug(f"GRACE_TIME: {greatest_grace_time}, {slurm_job_id}")
            await remove_job_by_slurm_job_id(slurm_job_id)

    await clean_jobs(squeue_result)


async def clean_jobs(squeue_result):
    """Clean the jobs that aren't running along with its bookings."""
    logger.debug("CLEAN_JOBS: start")

    cluster_jobs = [job.slurm_job_id for job in await get_cluster_jobs_from_backend()]
    if squeue_result is None:
        squeue_result = []

    jobs_not_running = [str(job["job_id"]) for job in squeue_result if job["state"] != "RUNNING"]
    all_jobs_squeue = [str(job["job_id"]) for job in squeue_result]
    logger.debug("CLEAN_JOBS: after building lists")

    delete_job_call = []

    for slurm_job_id in cluster_jobs:
        if slurm_job_id in jobs_not_running or slurm_job_id not in all_jobs_squeue:
            delete_job_call.append(remove_job_by_slurm_job_id(slurm_job_id))

    if not delete_job_call:
        logger.debug("CLEAN_JOBS: no need to clean")
        return

    await asyncio.gather(*delete_job_call)


async def create_or_update_reservation(reservation_data):
    """
    Create the reservation if it doesn't exist, otherwise update it.
    If the reservation cannot be updated, delete it and create a new one.
    """
    reservation = await scontrol_show_reservation()

    if reservation:
        updated = await scontrol_update_reservation(reservation_data, "30:00")
        if not updated:
            deleted = await scontrol_delete_reservation()
            LicenseManagerReservationFailure.require_condition(
                deleted, "Could not update or delete reservation."
            )
    else:
        created = await scontrol_create_reservation(reservation_data, "30:00")
        LicenseManagerReservationFailure.require_condition(created, "Could not create reservation.")


async def reconcile():
    """Generate the report and reconcile the license feature token usage."""
    logger.debug("Starting reconciliation")

    # Get cluster data
    configurations = await get_cluster_configs_from_backend()

    # Get feature bookings sum
    all_features_bookings_sum = await get_all_features_bookings_sum()

    # Get license usage from the cluster
    all_features_used_value = await get_all_features_used_values()

    # Delete bookings for jobs that reached the grace time
    logger.debug("Cleaning jobs by grace time")
    await clean_jobs_by_grace_time()
    logger.debug("Jobs cleaned by grace time")

    # Generate report and update the backend
    logger.debug("Reconciling licenses in the backend")
    license_usage_info = await update_features()
    logger.debug("Backend licenses reconciliated")

    reservation_data = []

    # Calculate how many licenses should be reserved for each license
    for license_data in license_usage_info:
        # Get license usage from license report
        product_feature = license_data.product_feature
        server_used = license_data.used
        total = license_data.total

        # Get booking information from backend
        booking_sum = all_features_bookings_sum[product_feature]

        # Get license server type and reserved from the configuration in the backend
        for configuration in configurations:
            for feature in configuration.features:
                if f"{feature.product.name}.{feature.name}" == product_feature:
                    license_server_type = configuration.type
                    reserved = feature.reserved

        # Get license usage from the cluster
        slurm_used = all_features_used_value[product_feature]

        """
        The reserved amount represents how many licenses are already in use:
        Either in the license server or booked for a job (bookings from other cluster as well).
        If the license has a reserved value, it should be reserved too.

        The reservation is not meant to be used by any user, it's a way to block usage of licenses
        """

        reservation_amount = server_used - slurm_used + booking_sum + reserved

        if reservation_amount < 0:
            reservation_amount = 0

        if reservation_amount > total:
            reservation_amount = total

        if reservation_amount:
            reservation_data.append(f"{product_feature}@{license_server_type}:{reservation_amount}")

    if reservation_data:
        logger.debug(f"Reservation data: {reservation_data}")

        # Create the reservation or update the existing one
        await create_or_update_reservation(",".join(reservation_data))
    else:
        logger.debug("No reservation needed")

    logger.debug("Reconciliation done")


async def update_features() -> List[LicenseReportItem]:
    """Send the license data collected from the cluster to the backend."""
    license_report = await report()

    if not license_report:
        logger.error(
            "No license data could be collected, check that tools are installed "
            "correctly and the right hosts/ports are configured in settings"
        )
        raise LicenseManagerEmptyReportError("Got an empty response from the license server")

    features_to_update = []

    for license in license_report:
        product, feature = license.product_feature.split(".")

        feature_data = {
            "product_name": product,
            "feature_name": feature,
            "total": license.total,
            "used": license.used,
        }

        features_to_update.append(feature_data)

    await make_feature_update(features_to_update)

    return license_report



"""
Parser for RLM
"""
import re
from typing import Optional

INT = r"\d+"
VERSION = rf"v{INT}\.{INT}"
HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTWORD2 = r"[a-zA-Z0-9-.]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD2})*"
FEATURE_NAME = r"[\w-]+"

FEATURE_LINE = rf"^\s*(?P<license_feature>{FEATURE_NAME}) {VERSION}$"
COUNT_LINE = r"^\s*count: (?P<count>\d+).*inuse: (?P<in_use>\d+).*$"
USAGE_LINE = rf"^\s*(?P<license_feature>{FEATURE_NAME}) {VERSION}: (?P<user_name>\w+)@(?P<lead_host>{HOSTNAME}) (?P<license_used_by_host>\d+).*$"  # noqa

RX_FEATURE = re.compile(FEATURE_LINE)
RX_COUNT = re.compile(COUNT_LINE)
RX_USAGE = re.compile(USAGE_LINE)


def parse_feature_line(line: str) -> Optional[str]:
    """
    Parse the feature line in the RLM output.
    Data we need:
    - ``feature``: license name
    """
    parsed_feature = RX_FEATURE.match(line)
    if parsed_feature is None:
        return None
    feature_data = parsed_feature.groupdict()
    if feature_data["license_feature"] == "rlmutil":
        return None

    return feature_data["license_feature"].lower()


def parse_count_line(line: str) -> Optional[dict]:
    """
    Parse the count line in the RLM output.
    Data we need:
    - ``count``: total amount of licenses
    - ``in_use``: quantity of licenses being use
    """
    parsed_count = RX_COUNT.match(line)
    if parsed_count is None:
        return None
    count_data = parsed_count.groupdict()

    return {
        "count": int(count_data["count"]),
        "in_use": int(count_data["in_use"]),
    }


def parse_usage_line(line: str) -> Optional[dict]:
    """
    Parse the usage line in the RLM output.
    Data we need:
    - ``feature``: license name
    - ``user_name``: user name
    - ``lead_host``: lead host
    - ``license_used_by_host``: quantity of licenses being use
    """
    parsed_usage = RX_USAGE.match(line)
    if parsed_usage is None:
        return None
    usage_data = parsed_usage.groupdict()

    return {
        "license_feature": usage_data["license_feature"].lower(),
        "user_name": usage_data["user_name"],
        "lead_host": usage_data["lead_host"],
        "booked": int(usage_data["license_used_by_host"]),
    }


def parse(server_output: str) -> dict:
    """
    Parse the output from the RLM server.
    Data we need:
    - ``feature``: license name
    - ``count``: total amount of licenses
    - ``in_use``: quantity of licenses being use
    """
    parsed_data: dict = {}
    feature_list: list = []

    for line in server_output.splitlines():
        feature = parse_feature_line(line)
        if feature:
            feature_list.append(feature)
            parsed_data[feature] = {"uses": [], "total": []}
            continue

        count = parse_count_line(line)
        if count:
            feature = feature_list[-1]
            parsed_data[feature] = {
                "used": count["in_use"],
                "total": count["count"],
                "uses": [],
            }
            continue

        usage = parse_usage_line(line)
        if usage:
            feature = usage["license_feature"]
            del usage["license_feature"]
            parsed_data[feature]["uses"].append(usage)
            continue

        else:
            continue

    return parsed_data
